from argparse import ArgumentParser, RawDescriptionHelpFormatter
from io import BytesIO
import json
from json import JSONDecodeError
from pathlib import Path
from zipfile import ZipFile

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as md
from matplotlib.ticker import FuncFormatter

from . import synth_format, __version__
from .rails import interpolate_spline


min_sec_formatter = FuncFormatter(lambda x,_: f"{int(x//60):d}:{int(x%60):02d}{'.{:03d}'.format(int((x%1)*1000)) if x%1 else ''}")

FINALIZED_BOOKMARK = {"time": 0.0, "name": "#smh_finalized"}
BEATMAP_JSON_FILE = "beatmap.meta.bin"
DIFFICULTIES = ("Easy", "Normal", "Hard", "Expert", "Master", "Custom")
NOTE_COLORS = {"right": "red", "left": "cyan", "single": "lime", "both": "gold"}
QUEST_WALL_DELAY = 0.051  # 51ms to guard against rounding errors 

def get_parser():
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        prog=f"python3 -m {__package__}.{Path(__file__).stem}",
        description='\n'.join([
            "Finalizes a map:"
            "  * Move certain walls up so they appear ingame as they do in the editor"
            "",
            "Also see the wiki on GitHub, which contains more detailed explainations, as well as some examples and images: https://github.com/adosikas/synth_mapping_helper/wiki",
        ]),
        epilog=f"Version: {__version__}",
    )
    parser.add_argument("input", type=Path, help="Input file")
    parser.add_argument("output", type=Path, help="Output file")
    parser.add_argument("--revert", action="store_true", help="Reverse finalizing options")
    parser.add_argument("--plot", action="store_true", help="Plot velocity and acceleration")
    return parser

def abort(reason: str):
    print("ERROR: " + reason)
    exit(1)

def main(options):
    if not options.input.is_file():
        abort("Input file is not a file, is the path correct?")
    out_buffer = BytesIO()  # buffer output zip file in memory, only write on success
    with ZipFile(options.input) as inzip, ZipFile(out_buffer, "w") as outzip:
        # copy all content except beatmap json
        outzip.comment = inzip.comment
        for info in inzip.infolist():
            if info.filename != BEATMAP_JSON_FILE:
                outzip.writestr(info, inzip.read(info.filename))
        # load beatmap json
        beatmap = json.loads(inzip.read(BEATMAP_JSON_FILE))
        finalized = FINALIZED_BOOKMARK in beatmap["Bookmarks"]["BookmarksList"]
        if not options.revert:
            if finalized:
                abort("Already finalized!")
        else:
            if not finalized:
                abort("Not finalized, will not revert!")
            beatmap["Bookmarks"]["BookmarksList"].remove(FINALIZED_BOOKMARK)
            
        # shift slides up so they appear ingame as they do in the editor
        for _, walls in beatmap["Slides"].items():
            for w in walls:
                if options.revert: 
                    w["position"][1] -= 2.1 * synth_format.GRID_SCALE
                else:
                    w["position"][1] += 2.1 * synth_format.GRID_SCALE
        if options.plot:
            bpm = beatmap["BPM"]
            for diff in DIFFICULTIES:
                notes_dict = beatmap["Track"][diff]
                if not notes_dict:
                    continue
                fig, axs = plt.subplots(5, 1, sharex=True, figsize=(16, 8))
                fig.suptitle(f"{beatmap['Author']} - {beatmap['Name']}: {diff}")
                (ax_x, ax_y, ax_vel, ax_acc, ax_walls) = axs
                ax_x.set_ylabel("X-Position (sq)")
                ax_x.set_ylim((8, -8))
                ax_x.grid(True)
                ax_y.set_ylabel("Y-Position (sq)")
                ax_y.set_ylim((-6, 6))
                ax_y.grid(True)

                ax_vel.set_ylabel("Velocity (sq/s)")
                ax_vel.set_ylim((0, 50))
                ax_acc.set_ylabel("Acceleration (sq/s²)")
                ax_acc.set_ylim((0, 50))

                ax_walls.set_ylabel("Walls")
                ax_walls.set_yticks([])

                ax_walls.set_xlabel("time (s)")
                ax_walls.xaxis.set_major_formatter(min_sec_formatter)
                # bookmarks
                for bookmark in beatmap["Bookmarks"]["BookmarksList"]:
                    time = int(bookmark["time"]) / 64 * 60 / bpm
                    for ax in axs:
                        ax.axvline(time, color="grey")
                    ax_x.text(time, 0.99, bookmark["name"], ha='left', va='bottom', rotation=45, transform=ax_x.get_xaxis_transform())
                # walls
                walls: dict[str, list[float]] = {w_type: [] for w_type in synth_format.WALL_TYPES}
                for slide in beatmap["Slides"][diff]:
                    walls[synth_format.WALL_LOOKUP[slide["slideType"]]].append(slide["time"] / 64 * 60 / bpm)
                walls["crouch"] = [w["time"] / 64 * 60 / bpm for w in beatmap["Crouchs"][diff]]
                walls["square"] = [w["time"] / 64 * 60 / bpm for w in beatmap["Squares"][diff]]
                walls["triangle"] = [w["time"] / 64 * 60 / bpm for w in beatmap["Triangles"][diff]]

                wall_markers = {
                    "wall_left": ("s", "left", 7),
                    "wall_right": ("s", "right", 6),
                    "angle_left": ("o", "left", 5),
                    "angle_right":  ("o", "right", 4),
                    "center": ("d", "full", 3),
                    "crouch": ("s", "top", 2),
                    "triangle": ("^", "none", 1),
                    "square": ("s", "none", 0),
                }
                hidden: list[tuple[str, float]] = []
                for wall_type, (marker, fill, y) in wall_markers.items():
                    last_time = None
                    for time in sorted(walls[wall_type]):
                        quest_hidden = last_time is not None and time - last_time < QUEST_WALL_DELAY
                        if quest_hidden:
                            print(f"Wall hidden on Quest: {wall_type} @ {int(time/60)}:{time%60:06.3f}, distance {int((time - last_time) * 1000)} ms")
                            hidden.append((wall_type, time))
                        else:
                            last_time = time
                            ax_walls.plot([time], [y], marker=marker, fillstyle=fill, color="green")
                # always draw hidden ones on top
                for wall_type, time in hidden:
                    marker, fill, y = wall_markers[wall_type]
                    ax_walls.plot([time], [y], marker=marker, fillstyle=fill, color="red")
                
                # notes & rails
                positions = [{} for _ in synth_format.NOTE_TYPES]
                for time in sorted(notes_dict):
                    note_list = notes_dict[time]
                    for note in note_list:
                        note_type, pos = synth_format.note_from_synth(bpm, 0, note)
                        color = NOTE_COLORS[synth_format.NOTE_TYPES[note_type]]
                        if pos.shape[0] != 1:
                            new_times = np.arange(pos[0,2], pos[-1,2], 1/64)
                            pos = interpolate_spline(pos, new_times)
                        pos[:, 2] *= 60 / bpm  # convert beat to second
                        for p in pos:
                            positions[note_type][p[2]] = p[:2]

                        # single note / rail head
                        ax_x.plot(pos[0, 2], pos[0, 0], color=color, marker="o")
                        ax_y.plot(pos[0, 2], pos[0, 1], color=color, marker="o")
                        if pos.shape[0] != 1:
                            # rail
                            ax_x.plot(pos[:, 2], pos[:, 0], color=color, marker="")
                            ax_y.plot(pos[:, 2], pos[:, 1], color=color, marker="")
                for note_type, poss in enumerate(positions):
                    if len(poss) <= 1:
                        continue
                    times = sorted(poss)
                    prev_t = times[0]
                    vels = []
                    accs = []
                    for t in times[1:]:
                        if t - prev_t < 1:
                            vel = (poss[t] - poss[prev_t])/(t-prev_t)
                            if vels and vels[-1][1] is not None:
                                accs.append((prev_t, vel - vels[-1][1]))
                            else:
                                accs.append((prev_t, None))
                            vels.append(((t+prev_t)/2, vel))
                        else:
                            vels.append(((t+prev_t)/2, None))
                        prev_t = t
                    ax_vel.plot([t for t, _ in vels], [np.sqrt(v.dot(v)) if v is not None else np.nan for _, v in vels], color=NOTE_COLORS[synth_format.NOTE_TYPES[note_type]])
                    ax_acc.plot([t for t, _ in accs], [np.sqrt(a.dot(a)) if a is not None else np.nan for _, a in accs], color=NOTE_COLORS[synth_format.NOTE_TYPES[note_type]])

                fig.tight_layout()
                plt.show()
                # fig.savefig(f"{diff}.png")
        if not options.revert:
            beatmap["Bookmarks"]["BookmarksList"].append(FINALIZED_BOOKMARK)
        # write modified beatmap json
        outzip.writestr(inzip.getinfo(BEATMAP_JSON_FILE), json.dumps(beatmap))
    # write output zip
    options.output.write_bytes(out_buffer.getbuffer())

if __name__ == "__main__":
    main(get_parser().parse_args())


import numpy as np

# generic helpers
def angle_to_xy(angles: "numpy array like") -> "numpy array like":
    """convert a angles in degrees to x and y coordinates"""
    rad_angles = np.radians(angles)
    return np.stack((np.cos(rad_angles), np.sin(rad_angles)), -1)

def random_ring(count: int) -> "numpy array (n, 2)":
    """random positons along a ring of radius 1"""
    return angle_to_xy(np.random.random_sample((count,)) * 360)

def random_xy(count: int, min: "numpy array (2)", max: "numpy array (2)") -> "numpy array (n, 2)":
    return np.random.random_sample((count, 2)) * (max - min) + min

# pattern generation
def spiral(
    fidelity: float, length: float, start_angle: float = 0
) -> "numpy array (n, 2)":
    """spiral with radius 1"""
    rot = np.arange(length) / fidelity  # in rotations, ie 1.0 = 360°
    if length % 1:
        # add a partial angle at the end
        rot = np.concatenate((rot, [(length / fidelity)]))
    return angle_to_xy(rot * 360 + start_angle)


def spikes(
    fidelity: float, length: float, start_angle: float = 0
) -> "numpy array (n, 2)":
    """spikes with radius 1, uses random when fidelity is 0"""
    output = np.zeros((int(length) * 3, 2))
    output[1::3] = (
        spiral(fidelity, length, start_angle) if fidelity else random_ring(int(length))
    )
    return output



from argparse import ArgumentParser, RawDescriptionHelpFormatter
import itertools
from json import JSONDecodeError
from pathlib import Path

import numpy as np

from . import synth_format, rails, pattern_generation, movement, __version__

_filter_groups = {
    "notes": synth_format.NOTE_TYPES,
    "walls": list(synth_format.WALL_TYPES),
    "slides": [name for name, (id, _) in synth_format.WALL_TYPES.items() if id < 100],
}

def _parse_number(val: str) -> float:
    if "/" in val:
        a, b = val.split("/",1)
        return float(a) / float(b)
    elif val.endswith("%"):
        return float(val[:-1]) / 100
    return float(val)

def _parse_range(val: str) -> tuple[float, float]:
    if ":" not in val:
        v = _parse_number(val)
        return (-v, v)
    split = val.split(":")
    if len(split) != 2:
        raise ValueError("Must be in the form 'max' or 'min:max'")
    try:
        min = _parse_number(split[0])
    except ValueError:
        raise ValueError("Error parsing minimum")
    try:
        max = _parse_number(split[1])
    except ValueError:
        raise ValueError("Error parsing maximum")
    return (min, max)

def _parse_xy_range(val: str) -> tuple[tuple[float, float], tuple[float, float]]:
    split = val.split(",")
    if len(split) != 2:
        raise ValueError("Must be in the form X_RANGE,Y_RANGE")
    try:
        x = _parse_range(split[0])
    except ValueError:
        raise ValueError("Error parsing x range")
    try:
        y = _parse_range(split[1])
    except ValueError:
        raise ValueError("Error parsing y range")
    return np.array((x, y)).transpose()


def _parse_position(val: str) -> tuple[float, float, float]:
    split = val.split(",")
    if len(split) != 3:
        raise ValueError("Must be in the form x,y,t")
    try:
        x = _parse_number(split[0])
    except ValueError:
        raise ValueError("Error parsing x")
    try:
        y = _parse_number(split[1])
    except ValueError:
        raise ValueError("Error parsing y")
    try:
        t = _parse_number(split[2])
    except ValueError:
        raise ValueError("Error parsing t")
    return (x,y,t)

def _movement_helper(data: synth_format.DataContainer, base_func, relative_func, pivot_func, relative: bool, pivot: list[int], *args, **kwargs) -> None:
    """pick the right function depending on relative or pivot being set"""
    if relative:
        data.apply_for_all(relative_func, *args, **kwargs)
    elif pivot is not None:
        data.apply_for_all(pivot_func, *args, pivot_3d=pivot, **kwargs)
    else:
        data.apply_for_all(base_func, *args, **kwargs)

def do_movement(options, data: synth_format.DataContainer, filter_types: list = synth_format.ALL_TYPES) -> None:
    if options.scale:
        _movement_helper(data, movement.scale, movement.scale_relative, movement.scale_from, options.relative, options.pivot, options.scale, types=filter_types)
    if options.rotate:
        _movement_helper(data, movement.rotate, movement.rotate_relative, movement.rotate_around, options.relative, options.pivot, options.rotate, types=filter_types)
    if options.offset:
        _movement_helper(data, movement.offset, movement.offset_relative, movement.offset, options.relative, options.pivot, options.offset, types=filter_types)
    if options.outset:
        _movement_helper(data, movement.outset, movement.outset_relative, movement.outset_from, options.relative, options.pivot, options.outset, types=filter_types)

def get_parser():
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        prog=f"python3 -m {__package__}.{Path(__file__).stem}",
        description='\n'.join([
            f"Note types:\n\t{', '.join(synth_format.NOTE_TYPES)}",
            f"Wall types:\n\t{', '.join(synth_format.WALL_TYPES)}",
            "",
            "Most number values accept decimals, percentages and fractions (ie '0.25',  '25%' or '1/4')",
            "If your value starts with a '-', you must add a = between option and value, ie '--rotate=-45' or '--offset=-1,0,0'",
            "",
            "Angles are in degrees",
            "Vectors are specified as 'x,y,time:'",
            "\tX/Y is measured in editor grid squares, where +x is right and +y is up",
            "\tTime is measured in measures/beats from the start of the selection",
            "",
            "Also see the wiki on GitHub, which contains more detailed explainations, as well as some examples and images: https://github.com/adosikas/synth_mapping_helper/wiki",
        ]),
        epilog=f"Version: {__version__}",
    )

    parser.add_argument("--use-original", action="store_true", help="When calling this multiple times, start over with orignal copied json")
    parser.add_argument("-f", "--filter-types", nargs="+", metavar="FILTER_TYPE", choices=synth_format.ALL_TYPES + tuple(_filter_groups), default=synth_format.ALL_TYPES, help=f"Only affect notes and walls of these types. Multiple types can be specified seperated by spaces, defaults to all")
    parser.add_argument("--invert-filter", action="store_true", help="Invert filter so everything *but* the filter is affected")

    preproc_group = parser.add_argument_group("pre-processing")
    preproc_group.add_argument("-b", "--bpm", type=_parse_number, help="Change BPM without changing timing")
    preproc_group.add_argument("--delete-others", action="store_true", help="Delete everything that doesn't match the filter")
    preproc_group.add_argument("--connect-singles", type=_parse_number, metavar="MAX_INTERVAL", help="Replace strings of single notes with rails if they are . Use with'--rails-to-singles=1' if you want to keep the singles")
    preproc_group.add_argument("--merge-rails", type=_parse_number, nargs="?", action="append", metavar="MAX_INTERVAL", help="Merge sequential rails. By default only joins rails that start close to where another ends (in X, Y AND time), but with MAX_INTERVAL only the time interval counts")
    preproc_group.add_argument("--swap-hands", action="store_true", help="Swap left and right hand notes.")
    preproc_group.add_argument("-n", "--change-notes", metavar="NEW_NOTE_TYPE", nargs="+", choices=synth_format.NOTE_TYPES, help=f"Change the type/color of notes. Specify multiple to loop over them.")
    preproc_group.add_argument("-w", "--change-walls", metavar="NEW_WALL_TYPE", nargs="+", choices=synth_format.WALL_TYPES, help=f"Change the type of walls. Specify multiple to loop over them.")

    rail_pattern_group = parser.add_argument_group("rail patterns")
    interp_group = rail_pattern_group.add_mutually_exclusive_group()
    interp_group.add_argument("--interpolate", type=_parse_number, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating using splines (similar to the game). Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
    interp_group.add_argument("--interpolate-linear", type=_parse_number, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating linearly. Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
    rail_pattern_group.add_argument("--shorten-rails", type=_parse_number, metavar="DISTANCE", help="Cut some distance from every rail. Supports fractions. When negative, cuts from the start instead of the end")
    rail_pattern_group.add_argument("--start-angle", type=_parse_number, default=0.0, metavar="DEGREES", help="Angle of the first node of the spiral in degrees. Default: 0/right")
    rail_pattern_group.add_argument("--radius", type=_parse_number, default=1.0, help="Radius of spiral or length of spikes")
    rail_op_group = rail_pattern_group.add_mutually_exclusive_group()
    rail_op_group.add_argument("--spiral", type=_parse_number, metavar="NODES_PER_ROT", help="Generate counterclockwise spiral around rails with this number of nodes per full rotation. Supports fractions. 2=zigzag, negative=clockwise")
    rail_op_group.add_argument("--spikes", type=_parse_number, metavar="NODES_PER_ROT", help="Generate spikes from rail, either spiraling (see --spiral) or random (when set to 0)")
    rail_pattern_group.add_argument("--spike-width", type=_parse_number, default=1/32, help="Width of spike 'base' in beats. Supports fractions. Should not be lower than 1/32 (the default) and should be lower than chosen interpolation interval")

    movement_group = parser.add_argument_group("movement", description="Operation order is always: scale, rotate, offset, outset")
    pivot_group = movement_group.add_mutually_exclusive_group()
    pivot_group.add_argument("-p", "--pivot", type=_parse_position, help="Pivot for outset, scale and rotate as x,y,t")
    pivot_group.add_argument("--note-pivot", choices=synth_format.NOTE_TYPES, help="Use position of first matching note as pivot (determined before any operations)")
    pivot_group.add_argument("--relative", action="store_true", help="Use first node of rails as pivot for outset, scale and rotate, and for walls offset because relative based on rotation")
    movement_group.add_argument("-s", "--scale", type=_parse_position, help="Scale positions by x,y,t. Use negative values to mirror across axis. Does NOT change the size of walls. Time-Scale of 2 means twice as long, not twice as fast.")
    movement_group.add_argument("-r", "--rotate", type=_parse_number, metavar="DEGREES", help="Rotate counterclockwise by this many degrees (negative for clockwise)")
    movement_group.add_argument("-o", "--offset", type=_parse_position, help="Move/Translate by x,y,t")
    movement_group.add_argument("--outset", type=_parse_number, metavar="DISTANCE", help="Move outwards (negative for inwards, can move 'across' pivot)")
    rail_stack_group = movement_group.add_mutually_exclusive_group()
    rail_stack_group.add_argument("--offset-along", choices=synth_format.NOTE_TYPES, help="Offset objects to follow notes and rails of the specified color")
    rail_stack_group.add_argument("--rotate-with", choices=synth_format.NOTE_TYPES, help="Rotate and outset the objects to follow notes and rails of the specified color")
    movement_group.add_argument("--offset-random", type=_parse_xy_range, metavar="[MIN_X:]MAX_X,[MIN_Y:]MAX_Y", help="Offset by a random amount in the X and Y axis. When no MIN is given, uses negative MAX.")

    movement_group.add_argument("-c", "--stack-count", type=int, help="Instead of moving, create copies. Must have time offset set.")

    postproc_group = parser.add_argument_group("post-processing")
    postproc_group.add_argument("--split-rails", action="store_true", help="Split rails at single notes")
    postproc_group.add_argument("--rails-to-singles", type=int, nargs="?", action="append", metavar="KEEP_RAIL", help="Replace rails with single notes at all nodes. KEEP_RAIL is optional and can be '1' if you want to keep the rail instead of replacing it")
    postproc_group.add_argument("--keep-alignment", action="store_true", help="Do NOT shift the start of selection to first element")

    return parser

def abort(reason: str):
    print("ERROR: " + reason)
    exit(1)

def main(options):
    try:
        data = synth_format.import_clipboard(options.use_original)
    except (JSONDecodeError, KeyError) as err:
        abort(f"Could not decode clipboard, did you copy somethinge else?\n\t{err!r}")

    # argument post-parsing
    if any(name in options.filter_types for name in _filter_groups):
        out_filters = []
        for inp in options.filter_types:
            if inp in _filter_groups:
                out_filters.extend(_filter_groups[inp])
            else:
                out_filters.append(inp)
        options.filter_types = out_filters
    if not options.invert_filter:
        # only have each entry once
        filter_types = list(set(options.filter_types))
    else:
        filter_types = list(
            t for t in synth_format.ALL_TYPES
            if t not in options.filter_types
        )
    if options.note_pivot:
        notes = getattr(data, options.note_pivot)
        if not notes:
            abort(f"Could not find any {options.note_pivot} notes")
        options.pivot = notes[sorted(notes)[0]][:3]

    # preprocessing
    if options.bpm:
        bpm_scale = [1, 1, options.bpm / data.bpm]
        data.apply_for_all(movement.scale, bpm_scale, types=filter_types)

    if options.delete_others:
        data = data.filtered(types=filter_types)
    if options.connect_singles:
        data.apply_for_note_types(rails.connect_singles, max_interval=options.connect_singles, types=filter_types)
    if options.merge_rails:
        if options.merge_rails[0] is None:
            data.apply_for_note_types(rails.merge_sequential_rails, types=filter_types)
        else:
            data.apply_for_note_types(rails.merge_rails, max_interval=options.merge_rails[0], types=filter_types)

    if options.swap_hands:
        data.left, data.right = data.right, data.left

    if options.change_notes:
        if len(options.change_notes) == 1:
            # to single type: just merge all dicts
            changed = {}
            for t in filter_types:
                if t in synth_format.NOTE_TYPES and t != options.change_notes[0]:
                    changed |= getattr(data, t)
                    setattr(data, t, {})
            # existing notes always have priority
            setattr(data, options.change_notes[0], changed | getattr(data, options.change_notes[0]))
        else:
            # to multiple types: cycle
            outputs = {}
            for t in options.change_notes:
                outputs[t] = {}
            for t in filter_types:
                if t in synth_format.NOTE_TYPES:
                    notes = getattr(data, t)
                    # note: looping happens independently for each note type
                    for time, new_type in zip(sorted(notes), itertools.cycle(options.change_notes)):
                        nodes = notes[time]
                        outputs[new_type][nodes[0,2]] = nodes
                    setattr(data, t, {})  # clear existing notes
            for t, notes in outputs.items():
                # existing notes (that did not get changed) always have priority
                setattr(data, t, notes | getattr(data, t))

    if options.change_walls:
        if len(options.change_walls) == 1:
            # to single type: just merge all arrays the same way
            new_type = synth_format.WALL_TYPES[options.change_walls[0]][0]
            def _change_wall_type(wall: "numpy array (1, 5)") -> "numpy array (1, 5)":
                new_wall = wall.copy()
                new_wall[..., 3] = new_type
                return new_wall
            data.apply_for_walls(_change_wall_type, types=filter_types)
        else:
            # to multiple types: cycle
            generator = itertools.cycle(synth_format.WALL_TYPES[t][0] for t in options.change_walls)
            def _change_wall_type(wall: "numpy array (1, 5)") -> "numpy array (1, 5)":
                new_wall = wall.copy()
                new_wall[..., 3] = next(generator)
                return new_wall
            data.apply_for_walls(_change_wall_type, types=filter_types)

    # rail patterns
    if options.interpolate:
        data.apply_for_notes(rails.interpolate_nodes, "spline", options.interpolate, types=filter_types)
    elif options.interpolate_linear:
        data.apply_for_notes(rails.interpolate_nodes, "linear", options.interpolate_linear, types=filter_types)
    if options.shorten_rails:
        data.apply_for_notes(rails.shorten_rail, options.shorten_rails, types=filter_types)
    if options.spiral:
        if (1 / options.spiral) % 1 == 0:
            abort("Chosen spiral factor divides 1 and would result in a straight rail. Refusing action!")
        def _add_spiral(nodes: "numpy array (n, 3)") -> "numpy array (n, 3)":
            nodes[:, :2] += pattern_generation.spiral(options.spiral, nodes.shape[0], options.start_angle) * options.radius
            return nodes
        data.apply_for_notes(_add_spiral, types=filter_types)
    if options.spikes is not None:
        def _add_spikes(nodes: "numpy array (n, 3)") -> "numpy array (n, 3)":
            count = nodes.shape[0]
            nodes = np.repeat(nodes, 3, axis=0)
            nodes[::3] -= options.spike_width
            nodes[1::3] -= options.spike_width/2
            nodes[:, :2] += pattern_generation.spikes(options.spikes, count, options.start_angle) * options.radius
            return nodes
        data.apply_for_notes(_add_spikes, types=filter_types)

    # movement
    if options.stack_count:
        if not options.offset or options.offset[2] == 0:
            abort("Cannot stack with time offset of 0")
        stacking = data.filtered(types=filter_types)
        if options.offset_along or options.rotate_with:
            if options.offset_along and options.rotate_with:
                abort("Cannot use offset-along and rotate-with at the same time")
            if options.relative:
                abort("Cannot use offset-along or rotate-with in relative mode")
            start_pos = rails.get_position_at(getattr(data, options.offset_along or options.rotate_with), 0)
            if start_pos is None:
                abort(f"No start position of {options.offset_along or options.rotate_with} notes found")
            if options.pivot is not None:
                start_pos -= options.pivot[:2]
        for i in range(options.stack_count):
            do_movement(options, stacking)
            tmp = stacking.filtered()  # deep copy
            if options.offset_along:
                cur_pos = rails.get_position_at(getattr(data, options.offset_along), (i+1) * options.offset[2])
                if cur_pos is None:
                    abort(f"No intermediate position of {options.offset_along} notes found at {(i+1) * options.offset[2]}")
                delta = cur_pos - start_pos
                tmp.apply_for_all(movement.offset, [delta[0], delta[1], 0])
            if options.rotate_with:
                cur_pos = rails.get_position_at(getattr(data, options.rotate_with), (i+1) * options.offset[2])
                if cur_pos is None:
                    abort(f"No intermediate position of {options.rotate_with} notes found at {(i+1) * options.offset[2]}")
                if options.pivot is not None:
                    cur_pos -= options.pivot[:2]
                angle_diff = np.degrees(np.arctan2(start_pos[0], start_pos[1])) - np.degrees(np.arctan2(cur_pos[0], cur_pos[1]))
                distance_diff = np.sqrt(cur_pos.dot(cur_pos)) - np.sqrt(start_pos.dot(start_pos))
                # outset all objects together by precalculating the offset based on rail position
                group_outset = [cur_pos[0], cur_pos[1], 0] / np.sqrt(cur_pos.dot(cur_pos)) * distance_diff
                if options.pivot is not None:
                    tmp.apply_for_all(movement.rotate_around, angle_diff, pivot_3d=options.pivot)
                else:
                    tmp.apply_for_all(movement.rotate, angle_diff)
                tmp.apply_for_all(movement.offset, group_outset)
            if options.offset_random is not None:
                random_offset = pattern_generation.random_xy(1, options.offset_random[0], options.offset_random[1])[0]
                tmp.apply_for_all(movement.offset, [random_offset[0], random_offset[1], 0])
            data.merge(tmp)
    else:
        if options.offset_along or options.rotate_with:
            if options.offset_along and options.rotate_with:
                abort("Cannot use offset-along and rotate-with at the same time")
            start_pos = rails.get_position_at(getattr(data, options.offset_along or options.rotate_with), 0)
            if start_pos is None:
                abort(f"No start position of {options.offset_along or options.rotate_with} notes found")
            if options.pivot is not None:
                start_pos -= options.pivot[:2]
        do_movement(options, data, filter_types=filter_types)
        if options.offset_along:
            cur_pos = rails.get_position_at(getattr(data, options.offset_along), options.offset[2])
            if cur_pos is None:
                abort(f"No intermediate position of {options.offset_along} notes found at {options.offset[2]}")
            if options.relative:
                abort("Cannot use offset-along or rotate-with in relative mode")
            delta = cur_pos - start_pos
            data.apply_for_all(movement.offset, [delta[0], delta[1], 0], types=filter_types)
        if options.rotate_with:
            cur_pos = rails.get_position_at(getattr(data, options.rotate_with), options.offset[2])
            if cur_pos is None:
                abort(f"No intermediate position of {options.rotate_with} notes found at {options.offset[2]}")
            if options.pivot is not None:
                cur_pos -= options.pivot[:2]
            angle_diff = np.degrees(np.arctan2(start_pos[0], start_pos[1])) - np.degrees(np.arctan2(cur_pos[0], cur_pos[1]))
            distance_diff = np.sqrt(cur_pos.dot(cur_pos)) - np.sqrt(start_pos.dot(start_pos))
            # outset all objects together by precalculating the offset based on rail position
            group_outset = [cur_pos[0], cur_pos[1], 0] / np.sqrt(cur_pos.dot(cur_pos)) * distance_diff
            if options.pivot is not None:
                data.apply_for_all(movement.rotate_around, angle_diff, pivot_3d=options.pivot, types=filter_types)
            else:
                data.apply_for_all(movement.rotate, angle_diff, types=filter_types)
            data.apply_for_all(movement.offset, group_outset, types=filter_types)
        if options.offset_random is not None:
            random_offset = pattern_generation.random_xy(1, options.offset_random[0], options.offset_random[1])[0]
            data.apply_for_all(movement.offset, [random_offset[0], random_offset[1], 0], types=filter_types)

    # postprocessing
    if options.split_rails:
        data.apply_for_note_types(rails.split_rails, types=filter_types)
    if options.rails_to_singles:
        data.apply_for_note_types(rails.rails_to_singles, keep_rail=bool(options.rails_to_singles[0]), types=filter_types)
    synth_format.export_clipboard(data, not options.keep_alignment)

if __name__ == "__main__":
    main(get_parser().parse_args())


#/usr/bin/env python3
from dataclasses import dataclass
import json

import numpy as np
import pyperclip

# For simplicity, we exclusively use grid coordinates (x, y) and use measures for time (z)
# These values are only needed to convert to / from the format the game uses

INDEX_SCALE = 64  # index = measure * INDEX_SCALE
GRID_SCALE = 0.1365  # xy_coord = xy_grid * GRID_SCALE

# Apparently the editor grid snap is off by a small amount...
X_OFFSET = 0.002
Y_OFFSET = 0.0012

# not sure where this comes from, but is needed to work with z coordinates
# z_coord = index * (bpm / BPM_DIVISOR)
# also maybe useful: bpm = (z_coord / index) * BPM_DIVISOR
BPM_DIVISOR = 1200

MS_PER_MIN = 60 * 1000

NOTE_TYPES = ("right", "left", "single", "both")
# Note: wall offsets are eyeballed
WALL_TYPES = {
    # name: (index, center_offset)
    # index 0-4 are the same "slideType" used in the JSON in the "slides" list
    "wall_right": (0, [4.05, -1.85]),
    "wall_left": (1, [-3.95, -1.85]),
    "angle_right": (2, [2.5, -2.2]),
    "center": (3, [0, -1.3]),
    "angle_left": (4, [-2.5, -2.2]),
    # 5-99: placeholder for future slide types
    # Since there can be only be one wall anyway, just treat these three the same as slides
    "crouch": (100, [0.1, 7.5]),
    "square": (101, [0, -2]),
    "triangle": (102, [0, -4.7]),
}
WALL_LOOKUP = {id: name for name, (id, _) in WALL_TYPES.items()}
WALL_OFFSETS = {id: np.array(offset + [0]) for _, (id, offset) in WALL_TYPES.items()}

ALL_TYPES = NOTE_TYPES + tuple(WALL_TYPES)

SINGLE_COLOR_NOTES = dict[float, list["numpy array (n, 3)"]]   # rail segment (n>1) and x,y,t
WALLS = dict[float, list["numpy array (1, 5)"]]    # x,y,t, type, angle

@dataclass
class DataContainer:
    original_json: str
    bpm: float
    right: SINGLE_COLOR_NOTES
    left: SINGLE_COLOR_NOTES
    single: SINGLE_COLOR_NOTES
    both: SINGLE_COLOR_NOTES
    walls: WALLS

    # Note: None of these functions are allowed to *modify* the dicts, instead they must create new dicts
    # This avoids requring deep copies for everything

    def apply_for_notes(self, f, *args, types: list = NOTE_TYPES, **kwargs) -> None:
        for t in types:
            if t not in NOTE_TYPES:
                continue
            notes = getattr(self, t)
            out = {}
            for _, nodes in sorted(notes.items()):
                out_nodes = f(nodes, *args, **kwargs)
                out[out_nodes[0, 2]] = out_nodes
            setattr(self, str(t), out)

    def apply_for_walls(self, f, *args, types: list = WALL_TYPES, **kwargs) -> None:
        wall_types = [WALL_TYPES[t][0] for t in types if t in WALL_TYPES]
        out_walls = {}
        for time_index, wall in sorted(self.walls.items()):
            if wall[0, 3] in wall_types:
                wall = f(wall, *args, **kwargs)
                out_walls[wall[0, 2]] = wall
            else:
                out_walls[time_index] = wall
        self.walls = out_walls

    def apply_for_all(self, f, *args, types: list = ALL_TYPES, **kwargs) -> None:
        self.apply_for_notes(f, *args, types=types, **kwargs)
        self.apply_for_walls(f, *args, types=types, **kwargs)

    # used when the functions needs access to all notes and rails of a color at one
    def apply_for_note_types(self, f, *args, types: list = NOTE_TYPES, **kwargs) -> None:
        for t in types:
            if t not in NOTE_TYPES:
                continue
            setattr(self, t, f(getattr(self, t), *args, **kwargs))

    def filtered(self, types: list = ALL_TYPES) -> "DataContainer":
        note_dicts = [
            getattr(self, t) if t in types else {}
            for t in NOTE_TYPES
        ]
        wall_types = [WALL_TYPES[t][0] for t in types if t in WALL_TYPES]
        wall_dict = {
            time_index: wall
            for time_index, wall in sorted(self.walls.items())
            if wall[0, 3] in wall_types
        }
        return DataContainer(self.original_json, self.bpm, *note_dicts, wall_dict)
        
    def merge(self, other: "DataContainer") -> None:
        for t in NOTE_TYPES:
            self_notes = getattr(self, t)
            other_notes = getattr(other, t)
            setattr(self, t, self_notes | other_notes)
        self.walls |= other.walls

# basic coordinate
def coord_from_synth(bpm: float, startMeasure: float, coord: list[float]) -> "numpy array (3)":
    return np.array([
        (coord[0] - X_OFFSET) / GRID_SCALE,
        (coord[1] - Y_OFFSET) / GRID_SCALE,
        # convert absolute coordinate to number of beats since start
        round((coord[2] * bpm / BPM_DIVISOR) * 64 - startMeasure) / 64,
    ])

def coord_to_synth(bpm: float, coord: "numpy array (3)") -> list[float]:
    return [
        (coord[0] * GRID_SCALE) + X_OFFSET,
        (coord[1] * GRID_SCALE) + Y_OFFSET,
        (coord[2] / bpm) * BPM_DIVISOR,
    ]

# full note dict
def note_from_synth(bpm: float, startMeasure: float, note_dict: dict) -> tuple[int, "numpy array (n, 3)"]:
    start = coord_from_synth(bpm, startMeasure, note_dict["Position"])
    note_type = note_dict["Type"]
    if note_dict["Segments"] is None:
        return note_type, start[np.newaxis]  # just add new axis
    else:
        return note_type, np.stack((start,) + tuple(coord_from_synth(bpm, startMeasure, node) for node in note_dict["Segments"]))

def note_to_synth(bpm: float, note_type: int, nodes: "numpy array (n, 3)") -> dict:
    return {
        "Type": note_type,
        "Position": coord_to_synth(bpm, nodes[0]),
        "Segments": [coord_to_synth(bpm, node) for node in nodes[1:]] if nodes.shape[0] > 1 else None,
    }

# full wall dict
def wall_from_synth(bpm: float, startMeasure: float, wall_dict: dict, wall_type: int) -> tuple[int, "numpy array (1, 5)"]:
    return np.concatenate((
        coord_from_synth(bpm, startMeasure, wall_dict["position"]) + WALL_OFFSETS[wall_type],
        (wall_type, wall_dict.get("zRotation", 0.0))
    ))[np.newaxis]

def wall_to_synth(bpm: float, wall: "numpy array (1, 5)") -> tuple[str, dict]:
    wall_type = int(wall[0, 3])
    pos = coord_to_synth(bpm, wall[0, :3] - WALL_OFFSETS[wall_type])
    wall_dict = {
        "time": round(wall[0, 2] * 64),
        "slideType": wall_type,
        "position": pos,
        "zRotation": wall[0, 4] % 360,  # note: crouch walls cannot be rotated, this will be ignored for them
        "initialized": True,  # no idea what this is for
    }
    # crouch, square and triangle are not in the "slides" list, each has their own list and they do not use the "slideType" key
    if wall_type < 100:  # we gave crouch, square and triangle the types 100, 101 and 102
        dest_list = "slides"
    else:
        dest_list = WALL_LOOKUP[wall_type] + "s"
        del wall_dict["slideType"]
    return dest_list, wall_dict

# full json
def import_clipboard(use_original: bool = False) -> DataContainer:
    original_json = pyperclip.paste()
    clipboard = json.loads(original_json)
    if "original_json" in clipboard:
        original_json = clipboard["original_json"]
    if use_original:
        clipboard = json.loads(original_json)
    bpm = clipboard["BPM"]
    startMeasure = clipboard["startMeasure"]
    # r, l, s, b
    notes: list[SINGLE_COLOR_NOTES] = [{} for _ in range(4)]
    for time_index, time_notes in clipboard["notes"].items():
        for note in time_notes:
            note_type, nodes = note_from_synth(bpm, startMeasure, note)
            notes[note_type][nodes[0,2]] = nodes

    walls: dict[float, list["numpy array (1, 5)"]] = {}
    # slides (right, left, angle_right, center, angle_right)
    for wall_dict in clipboard["slides"]:
        wall = wall_from_synth(bpm, startMeasure, wall_dict, wall_dict["slideType"])
        walls[wall[0, 2]] = wall
    # other (crouch, square, triangle)
    for wall_type in ("crouch", "square", "triangle"):
        for wall_dict in clipboard[wall_type + "s"]:
            wall = wall_from_synth(bpm, startMeasure, wall_dict, WALL_TYPES[wall_type][0])
            walls[wall[0, 2]] = wall

    return DataContainer(original_json, bpm, *notes, walls)

def export_clipboard(data: DataContainer, realign_start: bool = True):
    clipboard = {
        "BPM": data.bpm,
        "startMeasure": 0,
        "startTime": 0,
        "lenght": 0,
        "notes": {},
        "effects": [],
        "jumps": [],
        "crouchs": [],
        "squares": [],
        "triangles": [],
        "slides": [],
        "lights": [],
        "original_json": data.original_json,
    }
    first = 99999
    last = -99999
    for note_type, notes in enumerate((data.right, data.left, data.single, data.both)):
        for time_index, nodes in notes.items():
            clipboard["notes"].setdefault(round(time_index * 64), []).append(note_to_synth(data.bpm, note_type, nodes))
            if nodes[0, 2] < first:
                first = nodes[0, 2]
            if nodes[-1, 2] > last:
                last = nodes[-1, 2]
    for _, wall in data.walls.items():
        if wall[0, 2] < first:
            first = wall[0, 2]
        if wall[0, 2] > last:
            last = wall[0, 2]
        dest_list, wall_dict = wall_to_synth(data.bpm, wall)
        clipboard[dest_list].append(wall_dict)

    if realign_start:
        # position of selection start in beats*64 
        clipboard["startMeasure"] = round(first * 64)
        # position of selection start in ms
        clipboard["startTime"] = first * MS_PER_MIN / data.bpm
        # length of the selection in milliseconds
        # and yes, the editor has a typo, so we need to missspell it too
        clipboard["lenght"] = last * MS_PER_MIN / data.bpm
    # always update length
    clipboard["lenght"] = (last - first) * MS_PER_MIN / data.bpm

    pyperclip.copy(json.dumps(clipboard))




import numpy as np

# Note: None of these functions are allowed to *modify* the input array instance. Returning the same array (if nothing needed to be changed) is allowed.

# movement
def offset(data: "numpy array (n, m)", offset_3d: "numpy array (3)", pivot_3d: None = None) -> "numpy array (n, 3+)":
    """translate positions"""
    # pivot is ignored, but exists so this can be used as pivot func aswell
    offset_nd = np.zeros((data.shape[-1]))
    offset_nd[...,:3] = offset_3d
    return data + offset_nd

def offset_relative(data: "numpy array (n, m)", offset_3d: "numpy array (3)") -> "numpy array (n, 3+)":
    """translate positions, use relative coordinates for walls"""
    if data.shape[-1] == 3:
        return offset(data, offset_3d)
    # calculate rot matrix for angle of every wall
    rad_ang = np.radians(np.atleast_1d(data[:, 4]))
    rot_matrix = np.rollaxis(np.array((
        (np.cos(rad_ang), np.sin(rad_ang)),
        (-np.sin(rad_ang), np.cos(rad_ang)),
    )), -1)
    offset_nd = np.zeros((data.shape[-1]))
    offset_nd[...,:2] = np.array(offset_3d[:2]).dot(rot_matrix)  # offset is rotated for each wall
    offset_nd[...,2] = offset_3d[2]  # t stays as-is
    return data + offset_nd
    
def outset(data: "numpy array (n, m)", outset_scalar: float) -> "numpy array (n, 3+)":
    """move positions outwards"""
    angles = np.arctan2(data[..., 1], data[..., 0])
    normalized = np.zeros(data.shape)
    normalized[..., 0] = np.cos(angles)
    normalized[..., 1] = np.sin(angles)
    return data + normalized * outset_scalar

def outset_from(data: "numpy array (n, m)", outset_scalar: float, pivot_3d: "numpy array (3)") -> "numpy array (n, 3+)":
    """move positions away from pivot"""
    pivot_nd = np.zeros((data.shape[-1]))
    pivot_nd[...,:3] = pivot_3d
    return outset(data - pivot_nd, outset_scalar) + pivot_nd

def outset_relative(data: "numpy array (n, m)", outset_scalar: float) -> "numpy array (n, 3+)":
    """move positions away from pivot"""
    if data.shape[0] == 1:
        # no effect on single notes / walls
        return data
    return outset(data - data[0], outset_scalar) + data[0]


def scale(data: "numpy array (n, 3+)", scale_3d: "numpy array (3)") -> "numpy array (n, 3+)":
    """scale positions relative to center and start of selection"""
    if scale_3d[2] == 0:
        raise ValueError("Cannot have 0 for time scale")
    scale_nd = np.ones((data.shape[-1]))
    scale_nd[...,:3] = scale_3d
    output = data * scale_nd
    if scale_nd[2] < 0:  # reverse order of elements
        output = output[::-1]
    return output


def scale_from(
    data: "numpy array (n, 3+)", scale_3d: "numpy array (3)", pivot_3d: "numpy array (3)"
) -> "numpy array (n, 3+)":
    """scale positions relative to pivot"""
    pivot_nd = np.ones((data.shape[-1]))
    pivot_nd[...,:3] = pivot_3d
    return scale(data - pivot_nd, scale_3d) + pivot_nd

def scale_relative(
    data: "numpy array (n, 3+)", scale_3d: "numpy array (3)"
) -> "numpy array (n, 3+)":
    """scale positions relative to first node"""
    if data.shape[0] == 1:
        # no effect on single notes / walls
        return data
    return scale(data - data[0], scale_3d) + data[0]

def rotate(data: "numpy array (n, 3+)", angle: float) -> "numpy array (n, 3+)":
    """rotate positions anticlockwise around center"""
    rad_ang = np.radians(angle)
    rot_matrix = np.identity(data.shape[-1])
    rot_matrix[:2,:2] = [
        [np.cos(rad_ang), np.sin(rad_ang)],
        [-np.sin(rad_ang), np.cos(rad_ang)],
    ]
    out = data.dot(rot_matrix)
    if data.shape[-1] >= 5:
        # just add to wall rotation
        out[..., 4] += angle
    return out

def rotate_around(
    data: "numpy array (n, 3+)", angle: float, pivot_3d: "numpy array (3)"
) -> "numpy array (n, 3+)":
    """rotate positions anticlockwise around pivot"""
    pivot_nd = np.zeros((data.shape[-1]))
    pivot_nd[...,:3] = pivot_3d
    return rotate(data - pivot_nd, angle) + pivot_nd

def rotate_relative(
    data: "numpy array (n, 3+)", angle: float
) -> "numpy array (n, 3+)":
    """rotate positions anticlockwise around first node/wall center"""
    if data.shape[0] == 1:
        if data.shape[-1] >= 5:
            # just add to wall rotation
            wall_rot = np.zeros((data.shape[-1]))
            wall_rot[4] = angle
            return data + wall_rot
        else:
            # no effect on single notes
            return data
    return rotate(data - data[0], angle) + data[0]



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='synth_mapping_helper',
    author='adosikas',
    author_email='',
    description='Toolbox for manipulating the JSON-Format used by Synth Riders Beatmap Editor in the clipboard',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/adosikas/synth_mapping_helper',
    project_urls={
        'Documentation': 'https://github.com/adosikas/synth_mapping_helper',
        'Bug Reports':
        'https://github.com/adosikas/synth_mapping_helper',
        'Source Code': 'https://github.com/adosikas/synth_mapping_helper',
        # 'Funding': '',
        # 'Say Thanks!': '',
    },
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    classifiers=[
        # see https://pypi.org/classifiers/
        'Development Status :: 3 - Alpha',

        'Intended Audience :: End Users/Desktop',
        'Topic :: Artistic Software',
        'Topic :: Games/Entertainment',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    install_requires=['numpy', 'scipy', 'matplotlib', 'pyperclip'],
    extras_require={
        'dev': ['check-manifest'],
        # 'test': ['coverage'],
    },
    # entry_points={
    #     'console_scripts': [  # This can provide executable scripts
    #         'run=synth_mapping_helper:main',
    # You can execute `run` in bash to run `main()` in src/synth_mapping_helper/__init__.py
    #     ],
    # },
)



import numpy as np
from scipy.interpolate import pchip_interpolate

from .synth_format import DataContainer, SINGLE_COLOR_NOTES

# How far last node and start note can be spaced for two rails to be merged
MERGE_ACCURACY_GRID = 1 / 8
MERGE_ACCURACY_BEAT = 1 / 64

def interpolate_linear(data: "numpy array (n, m)", new_z: "numpy array (x)") -> "numpy array (x, 3)":
    return np.stack((
        np.interp(new_z, data[:, 2], data[:, 0]),
        np.interp(new_z, data[:, 2], data[:, 1]),
        new_z,
    ), axis=-1)

def interpolate_spline(data: "numpy array (n, m)", new_z: "numpy array (x)") -> "numpy array (x, 3)":
    # add points in straight line from start and end to match shape more closely
    padded_data = np.concatenate(([2*data[0]-data[1]], data, [2*data[-1]-data[-2]]))
    return np.stack((
        pchip_interpolate(padded_data[:, 2], padded_data[:, 0], new_z),
        pchip_interpolate(padded_data[:, 2], padded_data[:, 1], new_z),
        new_z,
    ), axis=-1)

def get_position_at(notes: SINGLE_COLOR_NOTES, beat: float, interpolate_gaps: bool = True) -> "numpy array (2)":
    # single note
    if beat in notes:
        return notes[beat][0, 0:2].copy()
    last_before = None
    # on rail, or between notes: interpolate
    for time in sorted(notes):
        nodes = notes[time]
        if time > beat:  # passed by target beat: interpolate between last and this
            if interpolate_gaps and last_before is not None:
                return interpolate_spline(np.concatenate((last_before, nodes)), beat)[0:2]
            break
        last_before = nodes
        if nodes.shape[0] == 1:  # ignore single nodes
            continue
        if nodes[0, 2] < beat:  # on rail: interpolate between nodes
            return interpolate_spline(nodes, beat)[0:2]

    # neither
    return None

# Note: None of these functions are allowed to *modify* the input dict instance. Returning the same dict (if nothing needed to be changed) is allowed.

def split_rails(notes: SINGLE_COLOR_NOTES) -> SINGLE_COLOR_NOTES:
    """splits rails at single notes"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}

    for time in sorted(notes):
        if current_rail_start is not None and time >= current_rail_end:
            current_rail_start = None
            current_rail_end = None
        nodes = notes[time]
        if nodes.shape[0] > 1:
            current_rail_start = nodes[0, 2]
            current_rail_end = nodes[-1, 2]
            out[time] = nodes
        elif current_rail_start is not None:
            previous_nodes = out[current_rail_start]
            # last index that is *before* or *at* the note
            last_index = np.argwhere(previous_nodes[:, 2] <= nodes[0, 2])[-1][0]

            # use the remainder of current rail for new rail
            # includes [last_index] that gets replaced by note to avoid concatenating
            remainder = previous_nodes[last_index:].copy()
            remainder[0] = nodes[0]
            out[time] = remainder

            if previous_nodes[last_index, 2] == time:
                # when there is a existing node at the current time, leave it be
                out[current_rail_start] = previous_nodes[: last_index + 1]
            else:
                # else replace the next one with note to finish the rail and include it
                previous_nodes[last_index + 1] = nodes[0]
                out[current_rail_start] = previous_nodes[: last_index + 2]

            current_rail_start = remainder[0, 2]
            current_rail_end = remainder[-1, 2]

    return out


def merge_sequential_rails(notes: SINGLE_COLOR_NOTES) -> SINGLE_COLOR_NOTES:
    """merges rails where end and start a very close"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}

    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1:  # ignore single nodes
            out[time] = nodes
            continue

        if (
            current_rail_end is not None
            and time <= current_rail_end + MERGE_ACCURACY_BEAT
        ):
            previous_nodes = out[current_rail_start]
            if np.isclose(
                previous_nodes[-1, 2], nodes[0, 2], atol=MERGE_ACCURACY_BEAT
            ) and np.allclose(
                previous_nodes[-1, :2], nodes[0, :2], atol=MERGE_ACCURACY_GRID
            ):
                # add to rail
                out[current_rail_start] = np.concatenate((previous_nodes[:-1], nodes))
                current_rail_end = nodes[-1, 2]
                # add single note where rail started
                out[time] = nodes[:1]
                continue
    
        current_rail_start = time
        current_rail_end = nodes[-1, 2]
        out[time] = nodes

    return out

def merge_rails(notes: SINGLE_COLOR_NOTES, max_interval: float) -> SINGLE_COLOR_NOTES:
    """merges rails"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}

    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1:  # ignore single nodes
            out[time] = nodes
            continue

        if (
            current_rail_end is None
            or time > current_rail_end + max_interval
        ):
            current_rail_start = time
            current_rail_end = nodes[-1, 2]
            out[time] = nodes
        else:
            previous_nodes = out[current_rail_start]
            # when start and end are close, remove end node of previous rail before joining
            if np.isclose(
                previous_nodes[-1, 2], nodes[0, 2], atol=MERGE_ACCURACY_BEAT
            ) and np.allclose(
                previous_nodes[-1, :2], nodes[0, :2], atol=MERGE_ACCURACY_GRID
            ):
                previous_nodes = previous_nodes[:-1]

            # add to rail
            out[current_rail_start] = np.concatenate((previous_nodes, nodes))
            current_rail_end = nodes[-1, 2]
            # add single note where rail started
            out[time] = nodes[:1]

    return out

def connect_singles(notes: SINGLE_COLOR_NOTES, max_interval: float) -> SINGLE_COLOR_NOTES:
    """Turn single notes into rails"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}
    for time in sorted(notes):
        nodes = notes[time]

        if nodes.shape[0] != 1:  # existing rails are ignored and reset 
            current_rail_start = None
            current_rail_end = None
            out[time] = nodes
            continue

        if (
            current_rail_end is None
            or time > current_rail_end + max_interval + MERGE_ACCURACY_BEAT
        ):
            current_rail_start = time
            current_rail_end = nodes[-1, 2]
            out[time] = nodes
        else:
            out[current_rail_start] = np.concatenate((out[current_rail_start], nodes))
            current_rail_end = nodes[-1, 2]
            # note: does not readd the single notes to the output dict

    return out

def interpolate_nodes(
    data: "numpy array (n, 3)", mode: "'spline' or 'linear'", interval: float
) -> "numpy array (n, 3)":
    """places nodes at defined interval along the rail, interpolating between existing nodes"""
    if data.shape[0] == 1:  # ignore single nodes
        return data
    new_z = np.arange(data[0, 2], data[-1, 2], interval)
    # when it does not distribute evenly, append end
    if not np.isclose(new_z[-1], data[-1, 2]):
        new_z = np.append(new_z, [data[-1, 2]])

    if mode == "spline":
        return interpolate_spline(data, new_z)
    elif mode == "linear":
        return interpolate_linear(data, new_z)
    else:
        raise RuntimeError("Invalid iterpolation mode")


def rails_to_singles(notes: SINGLE_COLOR_NOTES, keep_rail: bool) -> SINGLE_COLOR_NOTES:
    """Turn all rails into single notes"""
    out: SINGLE_COLOR_NOTES = {}
    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1:  # ignore single nodes
            out[time] = nodes
            continue
        if keep_rail:
            out[nodes[0,2]] = nodes
        for node in nodes[int(keep_rail):]:  # when keeping the rail, don't overwrite the start with a single note
            out[node[2]] = node[np.newaxis]

    return out

def shorten_rail(
    data: "numpy array (n, 3)", distance: float
) -> "numpy array (n, 3)":
    """Cut a bit of the end or start of the rail"""
    if data.shape[0] == 1:  # ignore single nodes
        return data
    if distance > 0:  # cut at the end
        new_z = data[-1,2] - distance
        last_index = np.argwhere(data[:, 2] < new_z)[-1][0]  # last node before new end
        return np.concatenate((data[:last_index+1], interpolate_spline(data, [new_z])))
    else:  # cut at the start
        new_z = data[0,2] - distance  # distance is negative to minus is correct here
        first_index = np.argwhere(data[:, 2] > new_z)[0][0]  # first node after new start
        return np.concatenate((interpolate_spline(data, [new_z]), data[first_index:]))
